local Vehicle = {}

local random = math.random
local abs = math.abs
local sqrt = math.sqrt
local sin = math.sin
local cos = math.cos
local pow = math.pow

function Vehicle:new( img_vehicle )
	local instance = {}
	setmetatable( instance, self )
	self.__index = self
	self._init( instance, img_vehicle )
	return instance
end

function Vehicle:_init( img_vehicle )
	
	self.heading = 0
	self.wander_angle = 0

	self.max_speed = 1
	self.max_force = self.max_speed / 10

	self.separation_radius = 24
	self.alignment_radius = 24
	self.attraction_radius = 64

	self.location_x = love.graphics.getWidth() / 2
	self.location_y = love.graphics.getHeight() / 2

	self.velocity_x = 0
	self.velocity_y = 0

	self.acceleration_x = 0
	self.acceleration_y = 0

	self.img = img_vehicle
	
	self.img_size_x = img_vehicle:getWidth()
	self.img_size_y = img_vehicle:getHeight()
	self.img_origin_x = self.img_size_x / 2
	self.img_origin_y = self.img_size_y / 2
	self.img_scale = .3
	self.img_shape_color = { 255, 0, 127 }

end

function Vehicle:draw( )
	love.graphics.setColor(self.img_shape_color)
	love.graphics.draw(
		self.img,
		self.location_x,
		self.location_y,
		self.heading,
		self.img_scale,
		self.img_scale,
		self.img_origin_x,
		self.img_origin_y
	)
end

function Vehicle:update( vehicles, dt )
	wander( self, dt )

	flock( self, vehicles, dt)

	stay_in_box( self, dt )
	--seek( self, dt, love.mouse.getX(), love.mouse.getY() )
end

function flock(self, vehicles, dt)
	
	local attract_sum_x = 0
	local attract_sum_y = 0
	local attract_count = 0
	local align_sum_x = 0
	local align_sum_y = 0
	local align_count = 0
	local separate_sum_x = 0
	local separate_sum_y = 0
	local separate_count = 0

	for k = 1, #vehicles do
		
		local distance = sqrt(pow(self.location_x - vehicles[k].location_x, 2) + pow(self.location_y - vehicles[k].location_y, 2))
		
		attract_sum_x, attract_sum_y, attract_count = attract( self, vehicles[k], distance, attract_sum_x, attract_sum_y, attract_count )
		align_sum_x, align_sum_y, align_count = align( self, vehicles[k], distance, align_sum_x, align_sum_y, align_count )
		separate_sum_x, separate_sum_y, separate_count = separate( self, vehicles[k], distance, separate_sum_x, separate_sum_y, separate_count )

	end
	
	apply_attraction(self, attract_sum_x, attract_sum_y, attract_count, dt)
	apply_aligment(self, align_sum_x, align_sum_y, align_count, dt)
	apply_separation(self, separate_sum_x, separate_sum_y, separate_count, dt)

end

function stay_in_box( self, dt )

	local steer_x = 0
	local steer_y = 0
	local box_limit = 16

	if (self.location_x < box_limit) then
	  desired_x = self.max_speed * 2 
	  desired_y = self.velocity_y
	  
	  steer_x = desired_x - self.velocity_x
	  steer_y = desired_y - self.velocity_y
	end

	if (self.location_x > love.graphics.getWidth() - box_limit) then
	  desired_x = - self.max_speed * 2 
	  desired_y = self.velocity_y
	  
	  steer_x = desired_x - self.velocity_x
	  steer_y = desired_y - self.velocity_y
	end

	if (self.location_y < box_limit) then
	  desired_x = self.velocity_x
	  desired_y = self.max_speed * 2
	  
	  steer_x = desired_x - self.velocity_x
	  steer_y = desired_y - self.velocity_y
	end

	if (self.location_y > love.graphics.getHeight() - box_limit) then
	  desired_x = self.velocity_x
	  desired_y = - self.max_speed * 2
	  
	  steer_x = desired_x - self.velocity_x
	  steer_y = desired_y - self.velocity_y
	end

	apply_force(self, dt, steer_x, steer_y);
end

function wander( self, dt )
	-- compute circle vector by cloning velocity vector
	local circle_x = self.velocity_x
	local circle_y = self.velocity_y
	
	-- normalize circle vector
	--local length = sqrt(pow(circle_x, 2) + pow(circle_y, 2))
	--if ( length > 0 ) then
	--	circle_x = circle_x / length
	--	circle_y = circle_y / length
	--end

	-- scale circle by scalar
	local circle_distance = 16
	circle_x = circle_x * circle_distance
	circle_y = circle_y * circle_distance

	-- create non-zero displacement vector 
	local displacement_x =  0
	local displacement_y = -1

	-- scale displacemment vector
	local circle_radius = circle_distance * 2
	displacement_x = displacement_x * circle_radius
	displacement_y = displacement_y * circle_radius

	-- compute wander displacement
	local length = sqrt((displacement_x * displacement_x) + (displacement_y * displacement_y))
	displacement_x = cos(self.wander_angle) * length
   	displacement_y = sin(self.wander_angle) * length

   	local angle_displacement = 1
	self.wander_angle = self.wander_angle + (random(0, 1) * angle_displacement) *100

   	-- create desired vector
   	local desired_x = circle_x + displacement_x
   	local desired_y = circle_y + displacement_y

	-- normalize desired vector
   	local length = sqrt(pow(desired_x,2) + pow(desired_y, 2))
	if ( length > 0 ) then
		desired_x = desired_x / length
		desired_y = desired_y / length
	end

   	-- max speed limitation
	desired_x = desired_x * self.max_speed
	desired_y = desired_y * self.max_speed

	-- compute steering force
	local steer_x = desired_x - self.velocity_x
	local steer_y = desired_y - self.velocity_y

   	apply_force( self, dt, steer_x, steer_y )
end

function separate(self, vehicle, distance, sum_x, sum_y, count)

	if ((distance > 0) and (distance < self.separation_radius)) then
	
		difference_x = self.location_x - vehicle.location_x
		difference_y = self.location_y - vehicle.location_y

		local length = sqrt(pow(difference_x, 2) + pow(difference_y, 2))
		if ( length > 0 ) then
			difference_x = difference_x / length
			difference_y = difference_y / length
		end

		difference_x = difference_x / distance
		difference_y = difference_y / distance

		sum_x = sum_x + difference_x
		sum_y = sum_y + difference_y

		count = count + 1
	
	end

	return sum_x, sum_y, count

end

function apply_separation(self, sum_x, sum_y, count, dt)

	if (count > 0)	then
		sum_x = sum_x / count
		sum_y = sum_y / count
	end

	length = sqrt(pow(sum_x, 2) + pow(sum_y, 2))
	if ( length > 0 ) then
		sum_x = sum_x / length
		sum_y = sum_y / length
	end

	sum_x = sum_x * self.max_speed
	sum_y = sum_y * self.max_speed

	local steer_x = sum_x - self.velocity_x
	local steer_y = sum_y - self.velocity_y
	
	apply_force(self, dt, steer_x, steer_y)

end

function align(self, vehicle, distance, sum_x, sum_y, count)

	if ((distance > 0) and (distance < self.alignment_radius)) then
	
		sum_x = sum_x + vehicle.velocity_x
		sum_y = sum_y + vehicle.velocity_y

		count = count + 1
	
	end

	return sum_x, sum_y, count

end

function apply_aligment(self, sum_x, sum_y, count, dt)
	
	if (count > 0) then
		
		-- average out the sum vector 
		sum_x = sum_x / count
		sum_y = sum_y / count

		-- normalize sum vector
		local length = sqrt(pow(sum_x, 2) + pow(sum_y, 2))
		if ( length > 0 ) then
			sum_x = sum_x / length
			sum_y = sum_y / length
		end

		-- scale sum vector by max speed
		sum_x = sum_x * self.max_speed
		sum_y = sum_y * self.max_speed

		-- compute steering force
		local steer_x = sum_x - self.velocity_x
		local steer_y = sum_y - self.velocity_y

		apply_force(self, dt, steer_x, steer_y)

	end

end

function attract(self, vehicle, distance, sum_x, sum_y, count)

	if ((distance > 0) and (distance < self.attraction_radius)) then
		sum_x = sum_x + vehicle.location_x
		sum_y = sum_y + vehicle.location_y

		count = count + 1
	end

	return sum_x, sum_y, count

end

function apply_attraction(self, sum_x, sum_y, count, dt)

	if (count > 0) then
		
		-- average out the sum vector 
		sum_x = sum_x / count
		sum_y = sum_y / count

		-- compute desired vector
		local desired_x = sum_x - self.location_x;
		local desired_y = sum_y - self.location_y;
		
		-- normalize sum vector
		local length = sqrt(pow(desired_x, 2) + pow(desired_y, 2))
		if ( length > 0 ) then
			desired_x = desired_x / length
			desired_y = desired_y / length
		end

		-- scale sum vector by max speed
		desired_x = desired_x * (self.max_speed / 2)
		desired_y = desired_y * (self.max_speed / 2)

		-- compute steering force
		local steer_x = desired_x - self.velocity_x
		local steer_y = desired_y - self.velocity_y

		apply_force(self, dt, steer_x, steer_y)
	end

end

function seek( self, dt, target_x, target_y )
	-- compute desired vector
	local desired_x = target_x - self.location_x;
	local desired_y = target_y - self.location_y;

	-- normalize desired vector
	local length = sqrt((desired_x * desired_x) + (desired_y * desired_y))
	if ( length > 0 ) then
		desired_x = desired_x / length
		desired_y = desired_y / length
	end

	-- max speed limitation
	desired_x = desired_x * self.max_speed
	desired_y = desired_y * self.max_speed
 
	-- compute steering force
	steer_x = desired_x - self.velocity_x
	steer_y = desired_y - self.velocity_y

	if(love.mouse.isDown(1)) then
		steer_x = -steer_x
		steer_y = -steer_y
	end

	if(love.mouse.isDown(2)) then
		steer_x = steer_x * 2
		steer_y = steer_y * 2
	end

	apply_force(self, dt, steer_x, steer_y)
end

function apply_force( self, dt, force_x, force_y )

	force_x = force_x * dt
	force_y = force_y * dt
    
    if(force_x > self.max_force) then force_x = self.max_force end
    if(force_y > self.max_force) then force_y = self.max_force end
	
	self.acceleration_x = self.acceleration_x + force_x
	self.acceleration_y = self.acceleration_y + force_y

    self.velocity_x = self.velocity_x + self.acceleration_x
    self.velocity_y = self.velocity_y + self.acceleration_y

    if(self.velocity_x > self.max_speed) then self.velocity_x = self.max_speed end
    if(self.velocity_y > self.max_speed) then self.velocity_y = self.max_speed end

    self.location_x = self.location_x + self.velocity_x
    self.location_y = self.location_y + self.velocity_y

	self.acceleration_x = self.acceleration_x * 0
	self.acceleration_y = self.acceleration_y * 0

	self.heading = - math.atan2( self.velocity_x, self.velocity_y )
	
	self.img_shape_color =
	{
		(math.abs((self.velocity_x / self.max_speed) * 255) + math.abs((self.velocity_y / self.max_speed) * 127)) * self.max_speed * 2,
		0,
		(math.abs((self.velocity_x / self.max_speed) * 127) + math.abs((self.velocity_y / self.max_speed) * 255)) * self.max_speed * 2,
	}


end

return Vehicle